--[[
##############################################
##-
##-  WakeSpams
##-  By Cyrila <Aspire> @ Bloodfeather-EU
##-
##-  Cooldowns For Your Raid
##-
##-
##-   * Copyright (C) 2010  Dan Jacobsen
##-   *
##-   * This program is free software: you can redistribute it and/or modify
##-   * it under the terms of the GNU General Public License as published by
##-   * the Free Software Foundation, either version 3 of the License, or
##-   * (at your option) any later version.
##-   *
##-   * This program is distributed in the hope that it will be useful,
##-   * but WITHOUT ANY WARRANTY; without even the implied warranty of
##-   * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
##-   * GNU General Public License for more details.
##-   *
##-   * You should have received a copy of the GNU General Public License
##-   * along with this program.  If not, see <http://www.gnu.org/licenses/>.
##-
##############################################
]]

local WakeSpams = LibStub("AceAddon-3.0"):NewAddon("WakeSpams", "AceConsole-3.0", "AceEvent-3.0", "AceTimer-3.0")
_G.WakeSpams = WakeSpams


-- ------------------------
--
-- ------------------------

local _G, type, tonumber, tostring = _G, type, tonumber, tostring;
local tinsert = tinsert;
local format, strsub, strfind, gsub = format, string.sub, string.find, string.gsub;
local ipairs = ipairs;
local ceil = math.ceil;

local UnitName, GetSpellLink, UnitExists, UnitGUID, GetUnitName = _G.UnitName, _G.GetSpellLink, _G.UnitExists, _G.UnitGUID, _G.GetUnitName;
local GetSpellLink, GetSpellInfo, GetTime = _G.GetSpellLink, _G.GetSpellInfo, _G.GetTime;
local IsInInstance = _G.IsInInstance;


-- --------------------------------------
--  INTERNAL VARIABLES
-- --------------------------------------

--[===[@debug@
local debug = true
--@end-debug@]===]

local interval = 0.6
local state = 1
local timer = {}
local LastSent = {0, nil}
local Queue = {}
local Active = {}
local FadedActive = {}
local Detect = {}
local LatestDestName = nil

local PLAYER = UnitGUID("player")
local PET = UnitGUID("pet")
local PLAYERNAME = UnitName("player")

local dbset
local dbsetmany


-- ------------------------
--	WRAPPERS
-- ------------------------

function WakeSpams:print(s)
	DEFAULT_CHAT_FRAME:AddMessage("|cFFFFCC00[WakeSpams]|r "..s)
end

function WakeSpams:debug(s)
	if(debug) then
		DEFAULT_CHAT_FRAME:AddMessage("|cFFFFCC00[WSDebug]|r "..s)
	end
end

function WakeSpams:error(s, l)
	error(s, l)
end

function WakeSpams:DBSetAttribute(k,v)
	WakeSpams.db.profile[k] = v
end
dbset = WakeSpams.DBSetAttribute

function WakeSpams:DBSetMultipleAttributes(tbl)
	for k,v in pairs(tbl) do
		WakeSpams:DBSetAttribute(k,v)
	end
end
dbsetmany = WakeSpams.DBSetMultipleAttributes


-- --------------------------------
--  DEFAULTS
-- --------------------------------

--[[ Version legend (ciphers):
	1=major wow version,
	2=minor wow version,
	3=release type(0 for debug, 1 for alpha, 2 for beta, 3 for release),
	4=major version,
	5=minor version,
	6=DATABASE structure revision 2nd cipher (of CURRENT releasetype-major-minor)
	7=DATABASE structure revision 1st cipher (of CURRENT releasetype-major-minor)
	
	e.g: database version 4133001 would be WoW patch 4.1, release, addon version 3.0 and the first database release for this version.
	
	I do this mainly as an easy way for myself to see exactly which database structure version it is in relation to wow patch etc because of the
	spell database of this addon and to provide easy conversion. Database revision number has nothing to do with the SVN revision number and is manually incremented depending
	on the changes I make to the database.
	
	I intend to make conversion data for all 3.x versions and keep them as similar to eachother as possible.
	Another reason for this change is to be less intrusive with the default spell database, which will only import settings to your own spell database if any spell
	has had specific changes.
]]
local dbversion = 4103002
local db
local defaults = {
	profile = {
		disabled = false,
		
		-- INCREASE THIS VALUE IF YOU HAVE ISSUES WITH AOE SPELL ANNOUNCING.
		-- Warning: Do not set to below 0.1
		delay = 0.2,
		
		replaceme = "me",
		replacetarget = "YOU",
		
		cacheEnabled = true, -- Cache spells cast by character
		spellCache = {},
		
		-- Note: Spelltypes have same args as spells in addition to their own. They are not all configurable from the in-game options yet.
		-- To modify check SpellList.lua for reference!
		spelltypeGlobal = {
			["SPELL_DISPEL"] = {
				title = "Dispels",
				disabled = true,
				activate = "SPELL_DISPEL",
				str = "spell:link Removed: spell:extra:link from spell:target",
			},
			["SPELL_DISPEL_FAILED"] = {
				title = "Dispel Resists",
				activate = "SPELL_DISPEL_FAILED",
				str = "spell:link Dispel Failed: spell:extra:link from spell:target",
			},
			["SPELL_INTERRUPT"] = {
				title = "Interrupts",
				activate = "SPELL_INTERRUPT",
				str = "spell:link Interrupted: spell:target spell:extra:link",
			},
		},
		
		spelltype = {
			["SPELL_AURA_APPLIED"] = {
				title = "Spell Buff/Debuff (Activated)",
				activate = "SPELL_AURA_APPLIED", -- Default aura activate/fade
				str = "Activated: spell:link spell:durations!",
				str_target = "Activated: spell:link on spell:target - spell:durations!",
			},
			["SPELL_AURA_REMOVED"] = {
				title = "Spell Buff/Debuff (Faded)",
				activate = "SPELL_AURA_REMOVED", -- Default aura activate/fade
				str = "Faded: spell:link",
				str_target = "Faded: spell:link",
			},
			["SPELL_MISSED"] = {
				title = "Spell Miss",
				activate = "SPELL_MISSED", -- Spell misses (Taunts in legacy WS)
				str = "Missed: spell:link on spell:target (spell:extra)"
			},
			["SPELL_RESURRECT"] = {
				title = "Spell Resurrection",
				activate = "SPELL_RESURRECT", -- Resurrections
				str = "spell:link on spell:target!",
			},
			["UNIT_SPELLCAST_START"] = {
				title = "Spell Cast (Start)",
				activate = "UNIT_SPELLCAST_START", -- Special event for spellcast start
				str = "Casting: spell:link on spell:target",
			},
			["SPELL_CAST_SUCCESS"] = {
				title = "Spell Cast (Successful)",
				activate = "SPELL_CAST_SUCCESS",
				str = "spell:link on spell:target!",
			},
			["SPELL_HEAL"] = {
				title = "Spell Heal",
				activate = "SPELL_HEAL",
				str = "spell:link healed spell:target for spell:amount!",
			},
		},
		
		spelldb = {
			[PLAYER] = {
			},
		},
		
		output = { -- If ChatType/Channel not available, say to SELF
			[1] = { -- Self
				"SELF", nil
			},
			[2] = { -- Party
				"PARTY", nil
			},
			[3] = { -- Raid
				"RAID", nil
			},
			[4] = { -- BG
				"WHISPER", "TARGET"
			},
			[5] = { -- WG/TB
				"WHISPER", "TARGET"
			},
			[6] = { -- Arena
				"PARTY", nil
			},
		},
	},
}


-- -----------------------
--  INITIALIZE ADDON
-- -----------------------

local utils = {}
local sdbm = {}

function WakeSpams:OnInitialize()
	self:print("!!!NOTICE!!! This is an early 3.0 alpha version. The config has not yet been updated and is therefore UNUSABLE.")
	self:print("Please help WakeSpams development by reporting bug(s) you find in this version on the ticket tracker at: http://www.wowace.com/addons/wakespams/tickets/.")
	
	-- Include modules
	utils = self.utils
	sdbm = self.sdbm
	
	-- Setup DB
	self.db = LibStub("AceDB-3.0"):New("WakeSpamsDB", defaults)
	db = self.db.profile
	
	-- Convert older databases
	if(db.version ~= dbversion) then
		self.db.profile = utils:ConvertDB(db.version or 0, db, defaults.profile)
	end
	
	-- Spell database
	--if not db.spelldb[PLAYER] then self.db.profile.spelldb[PLAYER] = {} end
	db.spelldb[PLAYER] = sdbm:InitSpellDB(db.spelldb[PLAYER] or {})
	
	db.version = dbversion
	
	-- Enable/disable
	if not db.disabled then
		self:SetEnabledState(true)
	else
		self:SetEnabledState(false)
	end
	
	-- Slash commands
	self:RegisterChatCommand("wakespams", "SlashHandler");
	self:RegisterChatCommand("ws", "SlashHandler");
	
	-- Static popup dialogs
	utils:RegisterPopups()
end

function WakeSpams:OnEnable()
	-- Register events
	self:RegisterEvent("COMBAT_LOG_EVENT_UNFILTERED");
	self:RegisterEvent("UNIT_SPELLCAST_SENT");
	self:RegisterEvent("UNIT_SPELLCAST_START");
	
	self:RegisterEvent("PLAYER_ENTERING_WORLD", "UpdateState");
	self:RegisterEvent("RAID_ROSTER_UPDATE", "Event_RosterUpdate");
	self:RegisterEvent("PARTY_MEMBERS_CHANGED", "Event_RosterUpdate");
	self:RegisterEvent("ZONE_CHANGED_NEW_AREA", "Event_ZoneChanged");
	
	-- Force update so the addon won't be stuck in a previous state before next roster update or zone change.
	-- (Also useful to force an update for whatever other reason the addon might be stuck).
	self:UpdateState()
end

function WakeSpams:OnDisable()
	-- Unregister events
	self:UnregisterEvent("COMBAT_LOG_EVENT_UNFILTERED");
	self:UnregisterEvent("UNIT_SPELLCAST_SENT");
	self:UnregisterEvent("UNIT_SPELLCAST_START");
	
	self:UnregisterEvent("PLAYER_ENTERING_WORLD");
	self:UnregisterEvent("RAID_ROSTER_UPDATE");
	self:UnregisterEvent("PARTY_MEMBERS_CHANGED");
	self:UnregisterEvent("ZONE_CHANGED_NEW_AREA");
end


-- ----------------------------
--  Config
-- ----------------------------

function WakeSpams:SlashHandler(input)
	self.Options:ShowConfig()
end


-- ------------------------------------------------------------
--  HANDLE COMBAT LOG AND SPELLCASTING EVENTS
-- ------------------------------------------------------------

local knownTypes = {"player","NPC","pet","vehicle"}

-- Possible Unit IDs
local knownUnitIDs = {
	"mouseover",
	"focus",
	"target",
	"player",
	"pet",
	"vehicle",
	"arena1",
	"arena2",
	"arena3",
	"arena4",
	"arena5",
	"boss1",
	"boss2",
	"boss3",
	"boss4",
}

local function getUnitID(GUID, name)
	-- Gets unit id from guid. Return name if no result
	for _,v in ipairs(knownUnitIDs) do
		if (GUID == UnitGUID(v)) then
			name = v
			break
		end
	end
	return name
end

local function testFlags(unitFlags, testFlags)
	return bit.band(unitFlags, testFlags) > 0
end

local function getUnitType(unitGUID)
	unitGUID = tonumber(unitGUID:sub(5,5),16)
	unitGUID = knownTypes[unitGUID % 8]
	return unitGUID
end


-- -------------------
--  Combat log parser
-- -------------------

function WakeSpams:COMBAT_LOG_EVENT_UNFILTERED(_, timestamp, event, _, sourceGUID, sourceName, sourceFlags, destGUID, destName, destFlags, spellid, spellname, _, arg1, arg2, ...)
	-- Test source
	if(sourceGUID ~= PLAYER and sourceGUID ~= PET) then return nil end
	
	-- Load spell into cache
	if(db.cacheEnabled and spellid) then
		if not db.spellCache[PLAYER] then
			db.spellCache[PLAYER] = {}
		end
		if not db.spellCache[PLAYER][spellid] then
			db.spellCache[PLAYER][spellid] = {
				spellName = spellname,
			}
		end
		if not db.spellCache[PLAYER][spellid][event] then
			db.spellCache[PLAYER][spellid][event] = true
		end
	end
	
	-- Prepare for local db
	local spell
	
	-- Prioritize global events
	if (db.spelltypeGlobal[event]) then
		spell = db.spelltypeGlobal[event]
		spell.isEvent = true
		if (spell.disabled) then return nil end
		
		-- Check event spell blacklist
		if (spell.blkSpell) then
			if (spell.blkSpell[spellid]) then return end
		end
	else
		-- The spell database does not exist. If so, why? Throw error
		assert(db.spelldb, "Critical Error: WakeSpams spell database went missing. Please forward this error log to the project manager(s).")
		
		-- Continue if the spell exists in the saved variables
		assert(db.spelldb[PLAYER], "Critical Error: WakeSpams spell database has not been intialized for this character. Please report this error If it remains after UI reload or application restart.")
		if not db.spelldb[PLAYER][spellid] then return nil end
		
		if(db.spelldb[PLAYER][spellid].detect) then
			-- Auto detect spell settings.
			if(event ~= "SPELL_CAST_FAILED") then
				self:SpellAutoDetect({timestamp, event, destGUID, destName, spellid, spellname}, true)
			end
			return true
		end
		
		-- Get the spell table from the spell list
		spell = db.spelldb[PLAYER][spellid]
		spell.dur       = spell.dur       or 0
		spell.countdown = spell.countdown or 0 --or spell.countdown
		
	end
	
	if (spell.disabled) then return nil end
	if not spell.events[event] then return nil end
	
	-- Get dest unit type
	local destUnitType = getUnitType(destGUID)
	
	-- Get dest unit (simple) reaction
	local destUnitIsFriend = testFlags(destFlags, COMBATLOG_OBJECT_REACTION_FRIENDLY)
	
	-- Test blacklisted units
	if (spell.blkUnits) then
		if (spell.blkUnits["hostile"] and not destUnitIsFriend) then return end
		if (spell.blkUnits["friendly"] and destUnitIsFriend) then return end
	end
	
	-- Get target UnitID
	local destUnitID = getUnitID(destGUID, destName)
	
	-- [ Ugly shit hack. TODO: Remove everything
	-- SPELL_AURA
	if (event == "SPELL_AURA_REMOVED") then
		if (spell.countdown and (spell.countdown > 0)) then
			-- Stop timers
			self:CancelTimer(timer[spellid].timer, true)
			timer[spellid] = nil
		end
		if not spell.fade then
			return nil
		end
		spell.faded = true
	elseif (event == "SPELL_AURA_APPLIED") then
		local name
		name, _, _, _, _, spell.dur, _, _, _, _, _ = UnitAura(destUnitID, spellname, nil, "PLAYER")
		if not name then name, _, _, _, _, spell.dur, _, _, _, _, _ = UnitAura(destUnitID, spellname, nil, "PLAYER|HARMFUL") end
		spell.dur = tonumber(("%d"):format(spell.dur))
		
		if (utils:isNumeric(spell.countdown) and spell.countdown > 0) then
			timer[spellid] = {}
			timer[spellid].elapsed = 0
			timer[spellid].timer = self:ScheduleRepeatingTimer("TimerHandle", 1, {spellid, spell.dur, spell.dur-spell.countdown, destName}) -- duration, duration-time to know when to start announcing
		end
	end
	-- End ]
	
	local msg
	
	-- Get announce format
	if (spell.hastarget) then
		msg = type(spell.events[event])=="string" and spell.events[event] or db.spelltype[event].str_target
	else
		msg = type(spell.events[event])=="string" and spell.events[event] or db.spelltype[event].str
	end
	
	-- Return nil if no message was formatted... Which shouldn't ever happen. Return error.
	assert(msg, ("Critical Error: WakeSpams message formatting failed. Spell ID: %s, Custom spell: %s, Default message: %s. Please forward this error log to the project manager(s)."):format(spellid,(spell.custom==true)and"YES"or"NO",(db.spelltype[event]==true)and(tostring(db.spelltype[event].str))or"Event not found."))
	
	-- -------------------------------
	-- From here: Announce Imminent.
	-- -------------------------------
	
	
	local msgtarget
	
	-- Format string
	msg = gsub(msg, "spell:link", function (v)
		local r = ""
		if(spellid~=32747) then
			r = GetSpellLink(spellid) or ""
		end
		return r
	end)
	msg = gsub(msg, "spell:name", spellname or "")
	msg = gsub(msg, "spell:duration", spell.dur or "")
	msg = gsub(msg, "spell:amount", tostring(arg1) or "")
	msg = gsub(msg, "spell:target", function(v)
		if(destName == UnitName("player") and db.replaceme) then
			v = db.replaceme
		else
			v, _ = UnitName(destUnitID) or destName
			v = utils:RaidIconText(destUnitID)..v
			msgtarget = v
		end
		return v
	end)
	msg = gsub(msg, "spell:extra:link", function(v)
		if(arg1 and (utils:isNumeric(arg1))) then
			v = GetSpellLink(arg1) or ""
		end
		return v
	end)
	msg = gsub(msg, "spell:extra", function(v)
		if(event == "SPELL_MISSED") then
			v = tostring(arg1) or ""
		else
			v = arg2 or ""
		end
		return v
	end)
	
	
	if(spell.aoe) then
		self:Enqueue(spellid, destName, msg, spell.dur, spell.faded)
		return true
	end
	if(spell.faded and spell.fademode) then
		self:FadedEnqueue({spellid, msg, destName, destGUID, spell.fademode, spell.fadearg}, true)
		return true
	end
	if(spell.fademode == 2) then
		self:FadedEnqueue(spell.fadearg, true)
	end
	
	
	-- Convert spellid for announce
	--spellid = (spell.isEvent) and event or spellid
	
	self:Announce(msg, (spell.isEvent) and event or spellid, destName, destUnitType=="player" and destUnitIsFriend, msgtarget)
end

function WakeSpams:UNIT_SPELLCAST_SENT(_, unit, _, _, destName)
	if(unit ~= "player" and unit ~= "pet") then return nil end
	-- Set the latest spell target
	LatestDestName = destName
end

function WakeSpams:UNIT_SPELLCAST_START(_, unit, spellname, arg1, arg2, spellid)
	if(unit ~= "player" and unit ~= "pet") then return nil end
	
	-- Credits to LibResComm and oRA2
	local target = LatestDestName
	if not target or target == UNKNOWN then
	
		if GameTooltipTextLeft1:IsVisible() then
			target = GameTooltipTextLeft1:GetText():match("^" .. CORPSE_TOOLTIP:replace("%s", "(.+)"))
		end
		
	end
	
	local targetGUID = UnitGUID(unit)
	
	-- Simulate combat log event
	self:COMBAT_LOG_EVENT_UNFILTERED(nil, nil, "UNIT_SPELLCAST_START", _, UnitGUID("player"), UnitName("player"), nil, targetGUID, target, nil, spellid, spellname)
end


-- ---------------------------
--  Spell Auto Detection
-- ---------------------------

function WakeSpams:SpellAutoDetect(a, notimer)
	-- 1 = timestamp, 2 = event, 3 = destGUID, 4 = destName, 5 = spellid, 6 = spellname
	StaticPopup_Hide("WS_AUTODETECTSPELL")
	if not notimer then
		local detect = Detect[a]
		detect.dur = (detect.dur) and detect.dur or 0
		detect.event = (detect.event == "SPELL_CAST_START") and "UNIT_SPELLCAST_START" or detect.event
		
		-- Default messages for events
		if(detect.event == "SPELL_HEAL" or detect.event == "SPELL_DAMAGE") then
			detect.txt = "Activated: spell:link on spell:target (spell:amount)"
		end
		if(detect.event == "UNIT_SPELLCAST_START") then
			detect.txt = "Casting: spell:link on spell:target"
		end
		
		self.db.profile.SpellDB[a] = {
			detect = nil,
			aoe = (detect.targets > 1) and true or nil,
			event = (detect.event ~= "SPELL_AURA_APPLIED") and detect.event or nil,
			dur = detect.dur,
			fade = (detect.dur >= 5) and true or nil,
			txt = detect.txt,
		}
		StaticPopup_Show("WS_DETECT_SUCCESS")
		Detect[a] = false
		return true
	end
	if not Detect[a[5]] then
		Detect[a[5]] = {
			starttime = a[1],
			destGUID = a[3] or "",
			event = a[2],
			targets = 0,
		}
		self:ScheduleTimer("SpellAutoDetect", 5, a[5])
	end
	if(a[2] == "SPELL_AURA_APPLIED") then
		local unit = getUnitID(a[3], a[4])
		local name, duration, expirationTime, spellId
		name, _, _, _, _, duration, expirationTime, _, _, _, spellId = UnitAura(unit, a[6], nil, "PLAYER")
		if not name then name, _, _, _, _, duration, expirationTime, _, _, _, spellId = UnitAura(unit, a[6], nil, "PLAYER|HARMFUL") end
		Detect[a[5]].dur = ceil(duration)
		if(spellId ~= a[5]) then self:print("Warning: Entered spell ID seems to be differing from ID found in Unit Auras. Recheck Spell ID!") end
		Detect[a[5]].event = a[2]
		Detect[a[5]].targets = Detect[a[5]].targets + 1
	else
		if(a[2] == "SPELL_CAST_FAILED") then
			return nil
		end
		Detect[a[5]].event = (Detect[a[5]].event == "SPELL_AURA_APPLIED") and Detect[a[5]].event or a[2]
	end
end


-- ----------------------------
--  Misc. event handlers
-- ----------------------------

function WakeSpams:Event_RosterUpdate()
	self:UpdateState()
end

function WakeSpams:Event_ZoneChanged()
	-- Quote (WoWWiki):
	-- "When this event fires, the UI may still think you're in the zone you just left. Don't depend on GetRealZoneText()
	-- and similar functions to report the new zone in reaction to ZONE_CHANGED_NEW_AREA. (untested for similar events)"
	-- Circumvented by adding a delay to the function
	self:ScheduleTimer("UpdateState", 0.1)
end

function WakeSpams:UpdateState()
	-- Announce state
	-- 1: Solo, 2: Party, 3: Raid, 4: Battleground, 5: WG and Tol Barad, 6: Arena
	--[===[@debug@
	self:debug("Updating the State Flag.")
	--@end-debug@]===]
	local _, instanceType = IsInInstance()
	if(instanceType == "arena") then
		state = 6
		--[===[@debug@
		self:debug("Set to: "..state)
		--@end-debug@]===]
		return true
	end
	if(GetNumRaidMembers() >= 1) then
		-- Credits to Kagaro for locale-independant TB/WG zone detection
		local aid = GetCurrentMapAreaID()
		if (aid == 708 or aid == 709) or (aid == 501 and not GetWintergraspWaitTime()) then
			-- your in Tol Brad (has 2 area markers) or WinterGrasp
			state = 5
			--[===[@debug@
			self:debug("Set to: "..state)
			--@end-debug@]===]
			return true
		end
		if(instanceType == "pvp") then
			state = 4
			--[===[@debug@
			self:debug("Set to: "..state)
			--@end-debug@]===]
			return true
		end
		
		state = 3
		--[===[@debug@
		self:debug("Set to: "..state)
		--@end-debug@]===]
		return true
	end
	if(GetNumRaidMembers() <= 0 and GetNumPartyMembers() >= 1) then
		state = 2
		--[===[@debug@
		self:debug("Set to: "..state)
		--@end-debug@]===]
		return true
	end
	state = 1
	--[===[@debug@
	self:debug("Set to: "..state)
	--@end-debug@]===]
	return true
end


-- ---------------------------
--  Special fade mode queue
-- ---------------------------

function WakeSpams:FadedEnqueue(args, notimer)
	-- spellid, msg, destName, destGUID, fademode, fadearg
	if(notimer) then
		if(utils:isNumeric(args)) then
			-- We got spellid, kill timer
			self:CancelTimer(FadedActive[args], true)
			return nil
		end
		if(args[5] == 1) then
			FadedActive[args[1]] = self:ScheduleTimer("FadedEnqueue", 0.2, args)
			return nil
		end
		return nil
	end
	local isplayer = UnitIsPlayer(getUnitID(args[4], args[3]))
	self:Announce(args[2], args[1], args[3], isplayer)
end


-- --------------------------------
--  AoE Spell Effect Queueing
-- --------------------------------

function WakeSpams:Enqueue(spellID, destName, msg, duration, faded)
	-- Workaround if for some reason an empty destName is returned from combatlog (?)
	if not spellID then return end
	if not destName then destName = "UNKNOWN" end
	
	if not Queue[spellID] and not Active[spellID] and not faded then
		-- Create new timer
		-- This is terribly programmed. Somebody shoot me and kill this tanglemess.
		self:ScheduleTimer("SendQueue", db.delay, {spellID, msg})
		duration = (duration >= 1) and duration or 60
		self:ScheduleTimer("SendFadedQueue", duration, {spellID, msg})
		Active[spellID] = 0
		Queue[spellID] = {}
	end
	if not faded then
		Active[spellID] = Active[spellID] + 1
		tinsert(Queue[spellID], {destName});
		return true
	end
	if(utils:isNumeric(Active[spellID])) then
		Active[spellID] = Active[spellID] - 1
		self:SendFadedQueue({spellID, msg}, true)
	end
end

function WakeSpams:SendFadedQueue(args, notimer)
	local spellID, msg = args[1], args[2]
	local spell = (db.SpellDB[spellid]) and db.SpellDB[spellid] or self.DefaultSpellDB[spellID]
	
	if not utils:isNumeric(Active[spellID]) then return false end
	if(notimer) then
		if(Active[spellID] >= 1) then return false end
	end
	
	if(spell.aoe and spell.fade) then
		-- Go!
		if(spell.txtend) then
			msg = spell.txtend
			msg = gsub(msg, "spell:link", GetSpellLink(spellID) or "")
		else
			msg = ("Faded: %s"):format(GetSpellLink(spellID))
		end
		self:Announce(msg, spellID)
	end
	
	Active[spellID] = nil
	Queue[spellID] = nil
	return true
end

function WakeSpams:SendQueue(args)
	local spellID, msg = args[1], args[2]
	local spell = (db.SpellDB[spellid]) and db.SpellDB[spellid] or self.DefaultSpellDB[spellID]
	if Queue[spellID] then
		local total
		local targetN
		for i,v in ipairs(Queue[spellID]) do
			total = i
			targetN = (targetN) and (targetN..", "..v[1]) or v[1]
		end
		if(msg and spell.txt) then
			msg = msg
		else
			if(total == 1) then
				msg = (("%s: Hit %d target! (%s)"):format(GetSpellLink(spellID),total,targetN))
			else
				msg = (("%s: Hit %d targets!"):format(GetSpellLink(spellID),total))
			end
		end
		--local msg = "SpellLink: Hit 3 targets (X, Y, Z)"
		self:Announce(msg, spellID)
	end
end


-- -------------------
--  Countdown Timer
-- -------------------

function WakeSpams:TimerHandle(args)
	-- args: 1 = spellid, 2 = d, 3 = ts, 4 = target
	timer[args[1]].elapsed = timer[args[1]].elapsed + 1
	
	if(timer[args[1]].elapsed >= args[3] and timer[args[1]].elapsed < args[2]) then
		self:Announce(""..args[2] - timer[args[1]].elapsed.."", args[1], args[4])
	end
end


-- -----------------------------
--  Communication
-- -----------------------------

function WakeSpams:Announce(msg, spellid, target, isplayer, msgtarget)
	if(GetTime() < LastSent[1]+interval and LastSent[2] == spellid and utils:isNumeric(spellid)) then -- Temporary chat throttle
		return nil
	end
	
	local tmpout = {}
	local outset
	local typeid, channel
	if(utils:isNumeric(spellid)) then
		outset = db.spelldb[PLAYER][spellid].output
		if(outset) then
			if(db.spelldb[PLAYER][spellid].whisper == true) then
				typeid, channel = "WHISPER", "TARGET"
			elseif(outset[state]) then
				typeid, channel = outset[state][1], outset[state][2]
			else
				typeid, channel = db.output[state][1], db.output[state][2]
			end
		else
			typeid, channel = db.output[state][1], db.output[state][2]
		end
	else
		outset = db.spelltypeGlobal[spellid].output
		if(outset) then
			if(outset[state]) then
				typeid, channel = outset[state][1], outset[state][2]
			else
				typeid, channel = db.output[state][1], db.output[state][2]
			end
		else
			typeid, channel = db.output[state][1], db.output[state][2]
		end
	end
	
	--- TODO: FIX HERE
	tinsert(tmpout, {typeid, channel})
	
	--[[
	if db.SpellDB[spellid] then
		if db.SpellDB[spellid].output then
			for i,v in ipairs(db.SpellDB[spellid].output) do
				tinsert(tmpout, {v[1], v[2]})
			end
		end
	end
	]]
	
	for i,v in ipairs(tmpout) do
		local msg = msg
		if(v[1]) then
		
			if(v[1] == "DISABLE") then
				return true
			end
		
			if(v[1] == "CHANNEL") then
				-- Resolve channel name
				v[2], _, _ = GetChannelName(v[2])
				if(v[2] == 0) then
					v[1] = "SELF"
					self:print("Custom channel name could not be resolved. You are not in that channel.")
				end
			end
			
			if(v[1] == "WHISPER" and v[2] == "TARGET") then
				-- Filters out if target is not a player
				if(isplayer) then
					v[2] = target
					if(db.replacetarget) then msg = gsub(msg, msgtarget, db.replacetarget) end
				else
					v[1] = "SELF"
					v[2] = nil
				end
			end
			
			--[===[@debug@
			self:debug("Announce spell to: "..v[1] or "nil"..", "..v[2] or "nil")
			--@end-debug@]===]
			
			self:ChatMSG(msg, v[1], nil, v[2])
		end
	end
	
	LastSent = {GetTime(), spellid}
end

function WakeSpams:ChatMSG(msg, mtype, mlang, marg)
	if(mtype=="SELF") then
		DEFAULT_CHAT_FRAME:AddMessage(msg)
	else
		if(mtype=="WHISPER" and marg==UnitName("player")) then
			DEFAULT_CHAT_FRAME:AddMessage(msg)
		else
			SendChatMessage(msg, mtype, mlang, marg)
		end
	end
end

